@@ -18,18 +18,12 @@ Perhaps best of all, a single Tdb file may contain one—or more—tables.
 import datetime
 import gzip
 import io
-import math
 import pathlib
 from xml.sax.saxutils import escape, unescape
 
 import editabletuple
 
-__version__ = '0.8.0'
-
-DATE_SENTINAL = datetime.date(1808, 8, 8)
-DATETIME_SENTINAL = datetime.datetime(1808, 8, 8, 8, 8, 8)
-INT_SENTINAL = -1808080808
-REAL_SENTINAL = -1808080808.0808
+__version__ = '0.9.0'
 
 
 class Tdb:
@@ -136,14 +130,19 @@ def _read_tdb(text):
 def _read_meta(text, lino):
     found, text, lino = _find(text, '%', 'expected to find "%"', lino)
     table = Table()
-    field_name = None
+    fieldname = None
     for i, part in enumerate(found.split()):
         if i == 0:
             table.name = part
         elif i % 2 != 0:
-            field_name = part
+            fieldname = part
         else:
-            table.fields_meta.append(MetaField(field_name, part))
+            null_allowed = False
+            if part.endswith('?'):
+                null_allowed = True
+                part = part[:-1]
+            table.fields_meta.append(MetaField(fieldname, part,
+                                               null_allowed=null_allowed))
     return text, table, lino + 1 # allow for %
 
 
@@ -166,8 +165,8 @@ def _read_records(text, table, lino):
             lino += 1
         elif c in ' \t\r': # ignore whitespace
             text = text[1:]
-        elif c == '!':
-            _handle_sentinal(field_meta, record, column, lino)
+        elif c == '?':
+            _handle_null(field_meta, record, column, lino)
             text, column = _advance(text, column)
         elif c in 'FfNn':
             _handle_bool(kind, False, record, column, lino)
@@ -226,12 +225,10 @@ def _advance(text, column):
     return text[1:], column + 1
 
 
-def _handle_sentinal(field_meta, record, column, lino):
-    sentinal = field_meta.sentinal
-    if sentinal is not None:
-        return sentinal
-    raise Error(
-        f'E140#{lino}:{field_meta.kind} fields don\'t allow sentinals')
+def _handle_null(field_meta, record, column, lino):
+    if not field_meta.null_allowed:
+        raise Error(E140.format(lino=lino, kind=field_meta.kind))
+    record[column] = None
 
 
 def _handle_bool(kind, value, record, column, lino):
@@ -325,10 +322,12 @@ def _find(text, what, message, lino):
 
 
 def _write_tdb(out, tables, decimals):
-    for table_name, table in tables.items():
-        out.write(f'[{table_name}')
+    for tablename, table in tables.items():
+        out.write(f'[{tablename}')
         for meta in table.fields_meta:
             out.write(f' {meta.name} {meta.kind}')
+            if meta.null_allowed:
+                out.write('?')
         out.write('\n%\n')
         for record in table.records:
             sep = ''
@@ -336,22 +335,23 @@ def _write_tdb(out, tables, decimals):
                 out.write(sep)
                 sep = ' '
                 kind = table.fields_meta[column].kind
-                if kind == 'bool':
+                if value is None:
+                    if table.fields_meta[column].null_allowed:
+                        out.write('?')
+                    else:
+                        raise Error(E240.format(kind=kind))
+                elif kind == 'bool':
                     out.write('T' if value else 'F')
                 elif kind == 'bytes':
                     out.write(f'({value.hex()})')
                 elif kind == 'date':
-                    out.write('!' if value == DATE_SENTINAL else
-                              value.isoformat())
+                    out.write(value.isoformat())
                 elif kind == 'datetime':
-                    out.write('!' if value == DATETIME_SENTINAL else
-                              value.isoformat()[:19])
+                    out.write(value.isoformat()[:19])
                 elif kind == 'int':
-                    out.write('!' if value == INT_SENTINAL else str(value))
+                    out.write(str(value))
                 elif kind == 'real':
-                    if math.isclose(value, REAL_SENTINAL):
-                        out.write('!')
-                    elif decimals <= 0:
+                    if decimals <= 0:
                         out.write(f'{value:g}')
                     else:
                         out.write(f'{value:.{decimals}f}')
@@ -363,36 +363,15 @@ def _write_tdb(out, tables, decimals):
 
 class MetaField:
 
-    def __init__(self, name, kind):
+    def __init__(self, name, kind, *, null_allowed=False):
         self.name = name
         self.kind = kind
-
-
-    @property
-    def default(self):
-        if self.kind == 'bool':
-            return False
-        if self.kind == 'bytes':
-            return b''
-        if self.kind == 'str':
-            return ''
-        return self.sentinal
-
-
-    @property
-    def sentinal(self):
-        if self.kind == 'date':
-            return DATE_SENTINAL
-        if self.kind == 'datetime':
-            return DATETIME_SENTINAL
-        if self.kind == 'int':
-            return INT_SENTINAL
-        if self.kind == 'real':
-            return REAL_SENTINAL
+        self.null_allowed = null_allowed
 
 
     def __repr__(self):
-        return f'{self.__class__.__name__}({self.name!r}, {self.kind!r})'
+        return (f'{self.__class__.__name__}({self.name!r}, {self.kind!r}, '
+                f'null_allowed={self.null_allowed!r})')
 
 
 class Table:
@@ -408,8 +387,7 @@ class Table:
     def RecordClass(self):
         if self._RecordClass is None:
             self._RecordClass = editabletuple.editabletuple(
-                self.name, *[field.name for field in self.fields_meta],
-                defaults=[field.default for field in self.fields_meta])
+                self.name, *[field.name for field in self.fields_meta])
         return self._RecordClass
 
 
@@ -431,6 +409,13 @@ class Error(Exception):
     pass
 
 
+E140 = ('E140#{lino}:{kind} fields don\'t allow nulls: provide a valid '
+        '{kind} or change the field\'s type to {kind}?')
+
+E240 = ('E240:can\'t write null to a not null field: provide a valid '
+        '{kind} or change the field\'s type to {kind}?')
+
+
 if __name__ == '__main__':
     import sys
     if len(sys.argv) == 1 or sys.argv[1] in {'-h', '--help'}:
